import os
from os import path
import re

from .slow_lzss import decompress

from . import dispatcher


def dump(binary, dest_dir):
    if not binary.startswith(b'<CHRP-BOOT>'): raise dispatcher.WrongFormat

    os.makedirs(dest_dir, exist_ok=True)

    a, b, c = binary.partition(b'</CHRP-BOOT>')
    chrp_boot = a + b
    if c.startswith(b'\r'): chrp_boot += b'\r'

    chrp_boot = chrp_boot.replace(b'\r', b'\n')

    # find the build-specific hex, and write out a clean version of the script
    chrp_boot_zeroed = bytearray(chrp_boot)
    constants = dict()
    for m in re.finditer(rb'h#\s+([A-Fa-f0-9]+)\s+constant\s+([-\w]+)', chrp_boot):
        key = m.group(2).decode('ascii')
        val = int(m.group(1), 16)
        constants[key] = val

        if key != 'elf-offset':
            for i in range(*m.span(1)):
                chrp_boot_zeroed[i:i+1] = b'0'

    with open(path.join(dest_dir, 'Bootscript'), 'wb') as f:
        f.write(chrp_boot_zeroed)

    if 'elf-offset' in constants:
        elf = binary[constants['elf-offset']:][:constants['elf-size']]
        dispatcher.dump(elf, path.join(dest_dir, 'MacOS.elf'))

    other_offset = constants.get('lzss-offset', constants.get('parcels-offset'))
    other_size = constants.get('lzss-size', constants.get('parcels-size'))
    parcels = binary[other_offset:][:other_size]

    if parcels.startswith(b'prcl'):
        filename = 'Parcels'
    else:
        filename = 'MacROM'
        parcels = decompress(parcels)

    dispatcher.dump(parcels, path.join(dest_dir, filename))



from .lowlevel import SuperMarioHeader, ConfigInfo

import struct
import shlex
import os
from os import path

from . import dispatcher


PAD = b'kc' * 100

HEADER_COMMENT = """
Automated dump of the ConfigInfo page of a Power Mac ROM
(at least one per ROM)

The first section contains the simple structure fields. The [LowMemory]
section instructs the kernel to set low-memory globals. The
[PageMappingInfo] section lists parts of ConfigInfo that tell the kernel
how to lay out the PowerPC page table. (Hint: lines not starting with a
tab are pointers into an array). The [BatMappingInfo] section similarly
tells the kernel how to lay out the Block Allocation Table registers.

Fields encoding the offset of a ROM component are computed from the base
of ConfigInfo, but for clarity are expressed here relative to the "BASE"
of ROM. If a second '=' is present, it indicates the name of a file to
insert at this location.
"""

HEADER_COMMENT = '\n'.join('# ' + l if l else '' for l in HEADER_COMMENT.strip().split('\n'))

CONFIGINFO_TEMPLATE = """
ROMImageBaseOffset=                                 # [28] Offset of Base of total ROM image
ROMImageSize=                                       # [2C] Number of bytes in ROM image
ROMImageVersion=                                    # [30] ROM Version number for entire ROM

                                                    #  ROM component Info (offsets are from base of ConfigInfo page)
Mac68KROMOffset=                                    # [34] Offset of base of Macintosh 68K ROM
Mac68KROMSize=                                      # [38] Number of bytes in Macintosh 68K ROM

ExceptionTableOffset=                               # [3C] Offset of base of PowerPC Exception Table Code
ExceptionTableSize=                                 # [40] Number of bytes in PowerPC Exception Table Code

HWInitCodeOffset=                                   # [44] Offset of base of Hardware Init Code
HWInitCodeSize=                                     # [48] Number of bytes in Hardware Init Code

KernelCodeOffset=                                   # [4C] Offset of base of NanoKernel Code
KernelCodeSize=                                     # [50] Number of bytes in NanoKernel Code

EmulatorCodeOffset=                                 # [54] Offset of base of Emulator Code
EmulatorCodeSize=                                   # [58] Number of bytes in Emulator Code

OpcodeTableOffset=                                  # [5C] Offset of base of Opcode Table
OpcodeTableSize=                                    # [60] Number of bytes in Opcode Table

                                    #  Offsets within the Emulator Data Page.
BootstrapVersion=                   # [64] Bootstrap loader version info
############## VALUES SPECIFIC TO THIS 68K EMULATOR VERSION ##############
BootVersionOffset=                  # [74] offset within EmulatorData of BootstrapVersion
ECBOffset=                          # [78] offset within EmulatorData of ECB
IplValueOffset=                     # [7C] offset within EmulatorData of IplValue

                                    #  Offsets within the Emulator Code.
EmulatorEntryOffset=                # [80] offset within Emulator Code of entry point
KernelTrapTableOffset=              # [84] offset within Emulator Code of KernelTrapTable

                                    #  Interrupt Passing Masks.
TestIntMaskInit=                    # [88] initial value for test interrupt mask
ClearIntMaskInit=                   # [8C] initial value for clear interrupt mask
PostIntMaskInit=                    # [90] initial value for post interrupt mask
##################### END OF EMULATOR-SPECIFIC VALUES ####################
LA_InterruptCtl=                    # [94] logical address of Interrupt Control I/O page
InterruptHandlerKind=               # [98] kind of handler to use

LA_InfoRecord=                      # [9C] logical address of InfoRecord page
LA_KernelData=                      # [A0] logical address of KernelData page
LA_EmulatorData=                    # [A4] logical address of EmulatorData page
LA_DispatchTable=                   # [A8] logical address of Dispatch Table
LA_EmulatorCode=                    # [AC] logical address of Emulator Code

                                    #  Address Space Mapping.
PageAttributeInit=                  # [B4] default WIMG/PP settings for PTE creation

                                    #  Only needed for Smurf
SharedMemoryAddr=                   # [35C] physical address of Mac/Smurf shared message mem

PA_RelocatedLowMemInit=             # [360] physical address of RelocatedLowMem

OpenFWBundleOffset=                                 # [364] Offset of base of OpenFirmware PEF Bundle
OpenFWBundleSize=                                   # [368] Number of bytes in OpenFirmware PEF Bundle

LA_OpenFirmware=                    # [36C] logical address of Open Firmware
PA_OpenFirmware=                    # [370] physical address of Open Firmware
LA_HardwarePriv=                    # [374] logical address of HardwarePriv callback
""".strip()


def extract_and_zero(binary, start, stop):
    ret = binary[start:stop]
    binary[start:stop] = bytes(stop - start)
    return ret


def find_configinfo(binary):
    # Find a ConfigIngo struct by checking every possible
    # place for a valid checksum. Ugly but quick.

    byte_lanes = [sum(binary[i::8]) for i in range(8)]

    for i in range(0, len(binary), 0x100):
        zeroed_byte_lanes = list(byte_lanes)
        for j in range(i, i+40):
            zeroed_byte_lanes[j % 8] -= binary[j]

        sum32 = [lane % (1<<32) for lane in zeroed_byte_lanes]

        sum64 = sum(lane << (k * 8) for (k, lane) in enumerate(reversed(zeroed_byte_lanes)))
        sum64 %= 1 << 64

        allsums = b''.join(x.to_bytes(4, byteorder='big') for x in sum32)
        allsums += sum64.to_bytes(8, byteorder='big')

        # if i == 0x30d000: # for figuring out pippin in future
        #     print(*['%02X' % x for x in allsums])
        #     print(*['%02X' % x for x in binary[i:i+len(allsums)]])

        if binary[i:i+len(allsums)] == allsums:
            break
    else:
        # Hack for Pippin ROM, which has bad checksum
        for i in range(0x300000, len(binary), 0x100):
            if binary[i+0x64:].startswith(b'Boot '):
                break
        else:
            return # failed!

    # Which structs share the BootstrapVersion signature?
    for j in range(0, len(binary), 0x100):
        if binary[i+0x64:i+0x74] == binary[j+0x64:j+0x74]:
            yield j


def dump_configinfo(binary, offset, filename_dict, push_line):
    s = ConfigInfo.unpack_from(binary, offset)

    push_line(HEADER_COMMENT + '\n')

    # First section (no [header]):
    # Raw key=value lines not resembling the struct in PCCInfoRecordsPriv.h
    for line in CONFIGINFO_TEMPLATE.split('\n'):
        if '=' in line:
            key, _, remainder = line.partition('=')
            raw_value = getattr(s, key)
            if key == 'InterruptHandlerKind':
                value = '0x%02X' % raw_value
            elif key == 'BootstrapVersion':
                value = shlex.quote(raw_value.decode('mac_roman'))
            elif key.endswith('Offset') and key.startswith(('Mac68KROM', 'ExceptionTable', 'HWInitCode', 'KernelCode', 'EmulatorCode', 'OpcodeTable', 'OpenFWBundle')):
                if raw_value == 0:
                    value = '0x00000000'
                else:
                    value = 'BASE0x%+X' % (raw_value - s.ROMImageBaseOffset)
            else:
                value = '0x%08X' % raw_value

            value = value.replace('0x-', '-0x').replace('0x+', '+0x')

            if key in filename_dict:
                value += '=' + shlex.quote(filename_dict[key])

            nuline = key + '=' + value
            while remainder.startswith('  ') and len(nuline) + len(remainder) > len(line):
                remainder = remainder[1:]
            nuline += remainder
            line = nuline

        push_line(line)

    push_line('')

    # Now dump the more structured parts of the ConfigInfo
    mapnames = ['sup', 'usr', 'cpu', 'ovl']

    segmaps = [[],[],[],[]]
    for i, blob in enumerate((s.SegMap32SupInit, s.SegMap32UsrInit, s.SegMap32CPUInit, s.SegMap32OvlInit)):
        for j in range(0, len(blob), 8):
            tpl = struct.unpack_from('>LL', blob, j)
            segmaps[i].append(tpl)

    def print_seg_ptrs_for_offset(segmap_offset):
        for header, list16 in zip(mapnames, segmaps):
            for seg_i, (seg_offset, seg_reg) in enumerate(list16):
                if seg_offset == segmap_offset:
                    push_line('segment_ptr_here=0x%X map=%s segment_register=0x%08X' % (seg_i, header, seg_reg))

    batmaps = [[],[],[],[]]
    for i, blob in enumerate((s.BatMap32SupInit, s.BatMap32UsrInit, s.BatMap32CPUInit, s.BatMap32OvlInit)):
        for j in reversed(range(0, 32, 4)):
            batmaps[i].append(((blob >> j) & 0xF) * 8)

    last_used_batmap = max(y for x in batmaps for y in x)

    def print_bat_ptrs_for_offset(batmap_offset):
        for header, list8 in zip(mapnames, batmaps):
            for bat_offset, bat_name in zip(list8, ['ibat0', 'ibat1', 'ibat2', 'ibat3', 'dbat0', 'dbat1', 'dbat2', 'dbat3']):
                if bat_offset == batmap_offset:
                    push_line('bat_ptr_here=%s map=%s' % (bat_name, header))

    lowmem = []
    lmoffset = s.MacLowMemInitOffset
    while any(binary[offset+lmoffset:][:4]):
        key, val = struct.unpack_from('>LL', binary, offset+lmoffset)
        lowmem.append((key, val))
        lmoffset += 8

    push_line('[LowMemory]')
    for key, val in lowmem:
        push_line('address=0x%08X value=0x%08X' % (key, val))
    push_line('')

    push_line('[PageMappingInfo]')
    if s.PageMapInitSize or any(s.SegMap32SupInit + s.SegMap32UsrInit + s.SegMap32CPUInit + s.SegMap32OvlInit):
        push_line('# Constants: PMDT_InvalidAddress = 0xA00, PMDT_Available = 0xA01')

        pagemapinit = binary[offset:][s.PageMapInitOffset:][:s.PageMapInitSize]

        for i in range(0, len(pagemapinit), 8):
            print_seg_ptrs_for_offset(i)

            pgidx, pgcnt, word2 = struct.unpack_from('>HHL', pagemapinit, i)
            attr = word2 & 0xFFF

            if attr == 0xA00:
                attr_s = 'PMDT_InvalidAddress'
            elif attr == 0xA01:
                attr_s = 'PMDT_Available'
            # elif attr & 
            else:
                attr_s = '0x%03X' % attr

            paddr = word2 >> 12
            if 'Rel' in attr_s:
                paddr_s = 'BASE+0x%05X' % ((paddr + offset) & 0xFFFFF)
            else:
                paddr_s = '0x%05X' % paddr

            if i == s.PageMapIRPOffset: push_line('special_pmdt=irp')
            if i == s.PageMapKDPOffset: push_line('special_pmdt=kdp')
            if i == s.PageMapEDPOffset: push_line('special_pmdt=edp')

            push_line('\tpmdt_page_offset=0x%04X pages_minus_1=0x%04X phys_page=%s attr=%s' % (pgidx, pgcnt, paddr_s, attr_s))

    push_line('')

    push_line('[BatMappingInfo]')
    if any(s.BATRangeInit) or s.BatMap32SupInit or s.BatMap32UsrInit or s.BatMap32CPUInit or s.BatMap32OvlInit:
        for i in range(0, len(s.BATRangeInit), 8):
            if i > last_used_batmap * 8: break

            print_bat_ptrs_for_offset(i)

            u, l = struct.unpack_from('>LL', s.BATRangeInit, i)

            is_relative = l & 0x200
            if is_relative:
                l = (offset + l) & 0xFFFFFFFF - is_relative

            bepi = u >> 17
            bl = (u >> 2) & 0x7FF
            vs = (u >> 1) & 1
            vp = u & 1

            brpn = l >> 17
            unk23 = (l >> 8) & 1
            wim = (l >> 4) & 0x7
            ks = (l >> 3) & 1
            ku = (l >> 2) & 1
            pp = l & 0x3

            bl_s = '0b' + bin(bl)[2:].zfill(6)
            wim_s = bin(wim)[2:].zfill(3)
            pp_s = bin(pp)[2:].zfill(2)

            if is_relative:
                brpn_s = 'BASE+0x%06X' % (brpn << 17)
            else:
                brpn_s = '0x%08X' % (brpn << 17)

            push_line('\tbepi=0x%08X bl_128k=%s vs=%d vp=%d brpn=%s unk23=%d wim=0b%s ks=%d ku=%d pp=0b%s' % (bepi << 17, bl_s, vs, vp, brpn_s, unk23, wim_s, ks, ku, pp_s))

    push_line('')



def is_powerpc(binary):
    return (len(binary) == 0x400000) and (PAD in binary[:0x300000])


def get_nk_version(nk):
    if nk.startswith(b'\x48\x00\x00\x0C'):
        # v2 NK has structured header
        return 'v%02X.%02X' % (nk[4], nk[5])

    for i in range(0, len(nk) - 8, 4):
        if nk[i:i+2] == b'\x39\x80': # li r12, ???
            if nk[i+4:i+8] == b'\xB1\x81\x0F\xE4': # sth r12, 0xFE4(r1)
                return 'v%02X.%02X' % (nk[i+2], nk[i+3]) # return the ???


def dump(orig_binary, dest_dir):
    if not is_powerpc(orig_binary): raise dispatcher.WrongFormat

    os.makedirs(dest_dir, exist_ok=True)

    # We will zero out parts as we go along extracting them
    binary = bytearray(orig_binary)

    ci_loc = []; ci_struct = [];
    for i in list(find_configinfo(binary)):
        ci_loc.append(i)
        ci_struct.append(ConfigInfo.unpack_from(binary, i))

        extract_and_zero(binary, i, i + 0x1000) # Keep it out of EverythingElse

    fields = []
    for field in ['Mac68KROM', 'ExceptionTable', 'HWInitCode', 'KernelCode', 'OpenFWBundle', 'ROMImageBase']:
        start = ci_loc[0] + ci_struct[0]._asdict()[field + 'Offset']
        if field == 'ROMImageBase': # This will contain everything not encompassed by 
            stop = len(binary)
        else:
            stop = start + ci_struct[0]._asdict()[field + 'Size']

        fields.append((start, stop, field))

    # Special case: the "EverythingElse" gets searched for last (emulator not yet extractable)
    fields = sorted(fields[:-1]) + fields[-1:]

    filename_dict = {} # Maps 'KernelCode' etc to a filename
    for start, stop, field in fields:
        # ConfigInfo is known to lie about these fields
        if field in 'HWInitCode KernelCode OpenFWBundle':
            stop = binary.find(bytes(1024), start)

        # Always grab a multiple of 4 bytes
        while stop % 4 != 0: stop += 1

        # Grab the fragment, and zero where it came from
        fragment = extract_and_zero(binary, start, stop)

        # Nothing to see here
        if len(fragment) == 0 or not any(fragment): continue

        filename = field.replace('Code', '').replace('Bundle', '').replace('Kern', 'NanoKern').replace('ROMImageBase', 'EverythingElse')
        if field == 'KernelCode':
            vers = get_nk_version(fragment)
            if vers: filename += '-' + vers

        filename_dict[field + 'Offset'] = filename

        dispatcher.dump(fragment, path.join(dest_dir, filename))

    # Finally, write out ConfigInfo with paths to the files that we create
    for i, cioffset in enumerate(ci_loc, 1):
        filename = 'Configfile-%d' % i
        with open(path.join(dest_dir, filename), 'w') as f:
            push_line = lambda x: print(x, file=f)
            dump_configinfo(orig_binary, cioffset, filename_dict, push_line)



from ast import literal_eval

def ascii_from_bytes(s):
    # convert to b'stuff' form, then cut off the first two and last one chars
    return ''.join(repr(bytes([c]))[2:-1] for c in s)

def bytes_from_ascii(s):
    # undo the above
    return b'"'.join(literal_eval('b"' + part + '"') for part in s.split('"')).decode('ascii')



# Thanks to Chris Warrick for script installation tips
# https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scripts/

import argparse
import sys
import os
from os import path
import shutil

from .slow_lzss import decompress

from . import dispatcher


def main(args=None):
    if args is None: args = sys.argv[1:]

    descriptions = {
        'dump': '''Break a ROM file into rebuildable parts. Any ROM
        released since the 660AV/840AV ("SuperMario") can be processed,
        including NewWorld ROMs. Because successive ROM formats tended
        to wrap layers around old ones, the image is dumped
        recursively.''',
        'build': '''Recreate a dumped ROM file. With minor exceptions,
        the result should be identical to the original. A NewWorld
        bootinfo file can be BinHex-encoded ('.hqx'), or have a '.idump'
        file created alongside.'''
    }

    for key in list(descriptions):
        descriptions[key] = ' '.join(descriptions[key].split())

    if not args or args[0] not in descriptions:
        print('usage: tbxi <command> [...]')
        print()
        print('The Mac OS Toolbox Imager')
        print()
        print('commands:')
        for k, v in descriptions.items():
            print('  ' + k.ljust(8) + ' ' + v.partition('.')[0])
        exit(1)

    command = args.pop(0)
    parser = argparse.ArgumentParser(prog='tbxi ' + command, description=descriptions[command])

    if command == 'dump':
        parser.add_argument('file', metavar='<input-file>', help='original file')
        parser.add_argument('-o', dest='output', metavar='<output-file>', help='destination (default: <input-file>.src)')
        args = parser.parse_args(args)

        if not args.output: args.output = args.file + '.src'

        with open(args.file, 'rb') as f:
            try:
                shutil.rmtree(args.output)
            except FileNotFoundError:
                pass

            dispatcher.dump(f.read(), args.output, toplevel=True)

    elif command == 'build':
        parser.add_argument('dir', metavar='<input-dir>', help='source directory')
        parser.add_argument('-o', dest='output', metavar='<output-file>', help='destination (default: Mac OS ROM)')
        args = parser.parse_args(args)

        if not args.output: args.output = 'Mac OS ROM'

        data = dispatcher.build(args.dir)

        if data.startswith(b'<CHRP-BOOT>'):
            base, ext = path.splitext(args.output)
            if ext.lower() == '.hqx':
                import binhex

                finfo = binhex.FInfo()
                finfo.Creator = b'chrp'
                finfo.Type = b'tbxi'
                finfo.Flags = 0

                bh = binhex.BinHex(('Mac OS ROM', finfo, len(data), 0), args.output)
                bh.write(data)
                bh.write_rsrc(b'')
                bh.close()

                return # do not write the usual way

            else:
                with open(args.output + '.idump', 'wb') as f:
                    f.write(b'tbxichrp')

        with open(args.output, 'wb') as f:
            f.write(data)


if __name__ == "__main__":
    main()



from os import path
import os
import shlex

from .lowlevel import SuperMarioHeader, ResHeader, ResEntry, FakeMMHeader, COMBO_FIELDS

from . import dispatcher


PAD = b'kc' * 100

HEADER_COMMENT = """
# Automated dump of Macintosh ROM resources

# The (optional) combo mask switches a resource based on the DefaultRSRCs
# field of the box's ProductInfo structure. (The low-memory variable at
# 0xDD8 points to ProductInfo, and the DefaultRSRCs byte is at offset
# 0x16.) The combo field is usually used for the Standard Apple Numeric
# Environment (SANE) PACKs 4 and 5.

# Summary of known combos:
# 0b01111000    AllCombos (DEFAULT)         Universal resource          
# 0b01000000    AppleTalk1                  Appletalk 1.0               
# 0b00100000    AppleTalk2                  Appletalk 2.0               
# 0b00110000    AppleTalk2_NetBoot_FPU      Has FPU and remote booting  
# 0b00001000    AppleTalk2_NetBoot_NoFPU    Has remote booting, no FPU  
# 0b00010000    NetBoot                     Has remote booting          

""".strip()

def clean_maincode(binary):
    binary = bytearray(binary)

    header = SuperMarioHeader.unpack_from(binary)
    modified_header = header._asdict()

    for k in list(modified_header):
        if k.startswith('CheckSum'):
            modified_header[k] = 0
    modified_header['RomRsrc'] = 0
    modified_header['RomSize'] = 1

    SuperMarioHeader.pack_into(binary, 0, **modified_header)

    return bytes(binary)


def is_supermario(binary):
    return (len(binary) in (0x200000, 0x300000)) and (PAD in binary)


def extract_decldata(binary):
    return binary[binary.rfind(PAD) + len(PAD):]


# Get a list of (entry_offset, data_offset, data_len)
def extract_resource_offsets(binary):
    # chase the linked list around
    offsets = []

    reshead = SuperMarioHeader.unpack_from(binary).RomRsrc
    link = ResHeader.unpack_from(binary, reshead).offsetToFirst
    while link:
        data = ResEntry.unpack_from(binary, link).offsetToData
        datasize = FakeMMHeader.unpack_from(binary, data - 16).dataSizePlus12 - 12
        offsets.append((link, data, datasize))
        link = ResEntry.unpack_from(binary, link).offsetToNext

    offsets.reverse()
    return offsets


def sanitize_macroman(binary):
    string = binary.decode('mac_roman')
    string = ''.join(c if c.isalpha() or c.isdigit() else '_' for c in string)
    return string


def express_macroman(binary):
    return repr(binary)[1:]
    accum = ''
    for b in binary:
        if b == ord(' '):
            accum += '\\ '
        elif b < 128 and chr(b).isprintable() and not chr(b).isspace():
            accum += chr(b)
        else:
            accum += '\\x%02X' % b
    return accum


def quodec(binary):
    return shlex.quote(binary.decode('mac_roman'))


def ljustspc(s, n):
    return (s + ' ').ljust(n)


def dump(binary, dest_dir):
    if not is_supermario(binary): raise dispatcher.WrongFormat

    os.makedirs(dest_dir, exist_ok=True)

    with open(path.join(dest_dir, 'Romfile'), 'w') as f:
        print(HEADER_COMMENT +  '\n', file=f)
        print('rom_size=%s\n' % hex(len(binary)), file=f)

        header = SuperMarioHeader.unpack_from(binary)

        main_code = clean_maincode(binary[:header.RomRsrc])
        dispatcher.dump(main_code, path.join(dest_dir, 'MainCode'))

        decldata = extract_decldata(binary)
        if decldata:
            dispatcher.dump(decldata, path.join(dest_dir, 'DeclData'))

        # now for the tricky bit: resources :(
        unavail_filenames = set(['', '.pef', '.pict'])
        types_where_Main_should_be_in_filename = set()

        known_forced_offsets = []

        for i, (hoffset, doffset, dlen) in enumerate(extract_resource_offsets(binary)):
            rsrc_dir = path.join(dest_dir, 'Rsrc')
            os.makedirs(rsrc_dir, exist_ok=True)

            data = binary[doffset:doffset+dlen]
            entry = ResEntry.unpack_from(binary, hoffset)

            offset_forced = False
            if hoffset < doffset: # Either offset was forced, or previous forced offset caused fit problems
                # Tricky guessing: check whether the data would have fit where the entry struct went
                for known in known_forced_offsets:
                    if hoffset <= known < hoffset + 16 + dlen:
                        break
                else:
                    offset_forced = True

            if offset_forced:
                known_forced_offsets.append(doffset - 16)

            # mmhead = FakeMMHeader.unpack_from(binary, doffset - 16)
            # assert mmhead.MagicKurt == b'Kurt'
            # assert mmhead.MagicC0A00000 == 0xC0A00000
            
            report_combo_field = COMBO_FIELDS.get(entry.combo, '0b' + bin(entry.combo >> 56)[2:].zfill(8))

            if entry.rsrcName == b'%A5Init':
                types_where_Main_should_be_in_filename.add(entry.rsrcType)

            # create a friendly ascii filename for the resource
            filename = '%s_%d' % (sanitize_macroman(entry.rsrcType), entry.rsrcID)
            if entry.rsrcName != b'Main' or entry.rsrcType in types_where_Main_should_be_in_filename:
                filename += '_' + sanitize_macroman(entry.rsrcName)
            if report_combo_field != 'AllCombos':
                filename += '_' + report_combo_field.replace('AppleTalk', 'AT')
            filename = filename.strip('_')
            while '__' in filename: filename = filename.replace('__', '_')
            if data.startswith(b'Joy!peff'): filename += '.pef'
            if entry.rsrcType == b'PICT': filename += '.pict'
            while filename in unavail_filenames: filename = '_' + filename

            unavail_filenames.add(filename)

            with open(path.join(rsrc_dir, filename), 'wb') as f2:
                f2.write(data)

            filename = path.join('Rsrc', filename)

            # Now, just need to dream up a data format
            report = ''
            report = ljustspc(report + 'type=' + quodec(entry.rsrcType), 12)
            report = ljustspc(report + 'id=' + str(entry.rsrcID), 24)
            report = ljustspc(report + 'name=' + quodec(entry.rsrcName), 48)
            report = ljustspc(report + 'src=' + shlex.quote(filename), 84)
            if report_combo_field != 'AllCombos':
                report = ljustspc(report + 'combo=' + report_combo_field, 0)
            if offset_forced:
                report = ljustspc(report + 'offset=0x%X' % (doffset - 16), 0)
            report = report.rstrip()

            print(report, file=f)



from collections import defaultdict, Counter
import os
from os import path
from shlex import quote
import struct

from . import dispatcher

from .slow_lzss import decompress
from .lowlevel import PrclNodeStruct, PrclChildStruct

def walk_tree(binary):
    """Get low level representation of tree

    e.g. [(prclnodetuple, [prclchildtuple, ...]), ...]
    """

    prclnode = None

    parents = []
    for i in iter(lambda: prclnode.link if prclnode else struct.unpack_from('>12xI', binary)[0], 0):
        prclnode = PrclNodeStruct.unpack_from(binary, offset=i)
        
        children = []
        for j in range(i + PrclNodeStruct.size, i + prclnode.hdr_size, prclnode.child_size):
            prclchild = PrclChildStruct.unpack_from(binary, offset=j)

            children.append(prclchild)

        parents.append((prclnode, children))

    return parents


def unique_binary_tpl(prclchild):
    return (prclchild.ptr, prclchild.packedlen, prclchild.compress)


def suggest_names_to_dump(parent, child, code_name):
    # We yield heaps of suggested filenames, and the shortest non-empty unique one gets chosen

    if parent.ostype == child.ostype == 'rom ':
        yield 'MacROM'
        return

    if 'AAPL,MacOS,PowerPC' in child.name and code_name == 'PowerMgrPlugin':
        if parent.a == 'cuda' and parent.b == 'via-cuda':
            yield 'PowerMgrPlugin.CUDA'
        elif parent.a == 'pmu' and parent.b == 'power-mgt':
            yield 'PowerMgrPlugin.PMU'
        elif parent.a == 'via-pmu-99' and parent.b == 'power-mgt':
            yield 'PowerMgrPlugin.PMU99'
        elif parent.a == 'via-pmu-2000' and parent.b == 'power-mgt':
            yield 'PowerMgrPlugin.PMU2000'
        elif parent.a == 'bal' and parent.b == 'power-mgt':
            yield 'PowerMgrPlugin.BlueBox'

    if ',' not in child.name: # All property names except driver,AAPL,MacOS,pef et al
        yield child.name

    if child.flags & 0x80: # special-node stuff
        yield child.name
        yield squish_name(child.name, parent.a, parent.b)

    if 'AAPL,MacOS,PowerPC' in child.name:
        if code_name:
            yield squish_name(code_name, parent.a, parent.b)
        else:
            yield squish_name(parent.a, parent.b)


def squish_name(*parts):
    squeeze = lambda x: x.lower().replace('-', '').replace('_', '')

    parts = list(parts)
    keepmask = [True] * len(parts)

    for i in range(len(parts)):
        for j in range(len(parts)):
            if i == j: continue
            if squeeze(parts[j]) == squeeze(parts[i]):
                if j > i: keepmask[j] = False
            elif squeeze(parts[j]) in squeeze(parts[i]):
                keepmask[j] = False

    truelist = []
    for i in range(len(parts)):
        if keepmask[i]: truelist.append(parts[i])

    return '.'.join(truelist)


def settle_name_votes(vote_dict):
    # Forbid duplicate names
    duplicate_names = set([''])
    for ka, va in vote_dict.items():
        for kb, vb in vote_dict.items():
            if ka is kb: continue

            for x in va:
                if x in vb:
                    duplicate_names.add(x)

    # Pick the shortest non-duplicate name
    decision = {}
    for k, v in vote_dict.items():
        allowed_names = [x for x in v if x not in duplicate_names]
        if allowed_names:
            decision[k] = min(allowed_names, key=len)

    return decision


def is_parcels(binary):
    return binary.startswith(b'prcl')


def dump(binary, dest_dir):
    if not binary.startswith(b'prcl'): raise dispatcher.WrongFormat

    os.makedirs(dest_dir, exist_ok=True)

    basic_structure = walk_tree(binary)

    # Decompress everything
    unpacked_dict = {}
    binary_counts = Counter()
    for prclnode, children in basic_structure:
        for prclchild in children:
            binary_counts[unique_binary_tpl(prclchild)] += 1

            data = binary[prclchild.ptr:prclchild.ptr+prclchild.packedlen]
            if prclchild.compress == 'lzss': data = decompress(data)

            unpacked_dict[unique_binary_tpl(prclchild)] = data

    # Suggest possible filenames for each blob
    name_vote_dict = defaultdict(list)
    for prclnode, children in basic_structure:
        # is there a prop that gives contextual name information?
        for check_child in children:
            if check_child.name == 'code,AAPL,MacOS,name':
                code_name = unpacked_dict[unique_binary_tpl(check_child)].rstrip(b'\0').decode('ascii')
                break
        else:
            code_name = None

        # now use that name to suggest names for all the children
        for prclchild in children:
            if prclchild.ostype in ('cstr', 'csta'): continue
            votes = suggest_names_to_dump(prclnode, prclchild, code_name)
            if unpacked_dict[unique_binary_tpl(prclchild)].startswith(b'Joy!'):
                votes = [v + '.pef' for v in votes]
            name_vote_dict[unique_binary_tpl(prclchild)].extend(votes)

    # Decide on filenames
    decision = settle_name_votes(name_vote_dict)

    # Dump blobs to disk
    for tpl, filename in decision.items():
        keep_this = True

        data = unpacked_dict[tpl]
        dispatcher.dump(data, path.join(dest_dir, filename))


    # Get printing!!!
    with open(path.join(dest_dir, 'Parcelfile'), 'w') as f:
        for prclnode, children in basic_structure:
            line = quote(prclnode.ostype)
            line += ' flags=0x%05x' % prclnode.flags
            if prclnode.a: line += ' a=%s' % quote(prclnode.a)
            if prclnode.b: line += ' b=%s' % quote(prclnode.b)

            print(line, file=f)

            for prclchild in children:
                line = '\t%s' % quote(prclchild.ostype)
                line += ' flags=0x%05x' % prclchild.flags
                if prclchild.name: line += ' name=%s' % quote(prclchild.name)

                if prclchild.ostype not in ('cstr', 'csta'):
                    filename = decision[unique_binary_tpl(prclchild)]
                    if prclchild.compress == 'lzss': filename += '.lzss'
                    line += ' src=%s' % filename

                if binary_counts[unique_binary_tpl(prclchild)] > 1:
                    line += ' deduplicate=1'

                print(line, file=f)

                if prclchild.ostype in ('cstr', 'csta'):
                    strangs = unpacked_dict[unique_binary_tpl(prclchild)].split(b'\0')[:-1]
                    for s in strangs:
                        line = '\t\t%s' % quote(s.decode('ascii'))

                        print(line, file=f)

            print(file=f)



import shlex
import ast
import struct
from os import path

from . import lowlevel
from . import dispatcher


ALIGN = 16
REV_COMBO_FIELDS = {v: k for (k, v) in lowlevel.COMBO_FIELDS.items()}


def checksum(binary):
    binary[:4] = bytes(4)
    binary[0x30:0x40] = bytes(16)

    lanes = [sum(binary[i::4]) & 0xFFFFFFFF for i in range(4)]
    struct.pack_into('>LLLL', binary, 0x30, *lanes)

    oneword = (sum(binary[::2])*256 + sum(binary[1::2])) & 0xFFFFFFFF
    struct.pack_into('>L', binary, 0, oneword)


def build(src):
    if not path.exists(path.join(src, 'Romfile')): raise dispatcher.WrongFormat

    romfile = dispatcher.build(path.join(src, 'Romfile')).decode('utf8').split('\n')

    rsrc_list = []
    for l in romfile:
        words = shlex.split(l, comments=True, posix=True)

        thisdict = {}
        for word in words:
            a, sep, b = word.partition('=')
            if sep == '=':
                if a == 'type':
                    b = b.encode('mac_roman')
                elif a in ('id', 'rom_size', 'offset'):
                    b = ast.literal_eval(b)

                thisdict[a] = b

        if 'rom_size' in thisdict:
            rom_size = thisdict['rom_size']
        elif 'type' in thisdict:
            rsrc_list.append(thisdict)

    rom = bytearray(b'kc' * (rom_size // 2))
    free_map = bytearray(b'X' * (rom_size // ALIGN))

    def rom_insert(offset, binary, letter=' '):
        letter = letter.encode('ascii')

        if offset + len(binary) > len(rom):
            raise IndexError('ROM too small to insert %r at %s' % (letter, hex(offset)))
        rom[offset:offset+len(binary)] = binary

        # Populate the informational array of letters
        # Only capitals can be overwritten
        start = offset // ALIGN
        stop = (offset + len(binary) - 1) // ALIGN + 1

        for i in range(start, stop):
            if free_map[i] != 20:
                if free_map[i:i+1].upper() != free_map[i:i+1]:
                    raise ValueError('Tried to insert %r over %r at %s' % (letter, free_map[i], hex(offset)))

            free_map[i:i+1] = letter

    def find_free(length):
        length = (length + ALIGN - 1) // ALIGN
        return free_map.index(b'X' * length) * ALIGN

    maincode = dispatcher.build(path.join(src, 'MainCode'))
    rom_insert(0, maincode, 'm')

    head_ptr = find_free(16)
    rom_insert(head_ptr, b'fake header', 'H')

    try:
        decldata = dispatcher.build(path.join(src, 'DeclData'))
    except FileNotFoundError:
        pass
    else:
        rom_insert(len(rom) - len(decldata), decldata, 'd')

    # now blat in the resources
    ent_ptr = 0
    bogus_off = 0x5C

    for r in rsrc_list:
        data = dispatcher.build(path.join(src, r['src']))

        # First place the data, including the fake MemMgr header
        if 'offset' in r:
            ofs = r['offset']
        else:
            ofs = find_free(16 + len(data))

        mm_ptr = ofs + 4
        data_ptr = ofs + 16

        # And insert that
        mm = lowlevel.FakeMMHeader.pack(
            MagicKurt=b'Kurt',
            MagicC0A00000=0xC0A00000,
            dataSizePlus12=len(data) + 12,
            bogusOff=bogus_off,
        )
        rom_insert(mm_ptr - 4, mm + data, 'r')

        # Calculate the independently-placed entry struct
        combo = r.get('combo', 'AllCombos')
        combo = REV_COMBO_FIELDS.get(combo, None)
        if combo is None:
            combo = ast.literal_eval(combo) << 56

        ent = lowlevel.ResEntry.pack(
            combo=combo,
            offsetToNext=ent_ptr, # this is the previous one
            offsetToData=data_ptr,
            rsrcType=r['type'],
            rsrcID=r['id'],
            rsrcAttr=0x58,
            rsrcName=r['name'].encode('mac_roman'),
        )
        # snip off pascal string padding
        ent = ent[:0x18 + ent[0x17]]

        # Place the entry struct
        ent_ptr = find_free(len(ent))
        rom_insert(ent_ptr, ent, 'e')

        bogus_off += 8

    head = lowlevel.ResHeader.pack(
        offsetToFirst=ent_ptr,
        maxValidIndex=4,
        comboFieldSize=8,
        comboVersion=1,
        headerSize=12,
    )
    rom_insert(head_ptr, head, 'h')

    # now set the size
    fields = lowlevel.SuperMarioHeader.unpack_from(rom)._asdict()
    fields['RomRsrc'] = head_ptr
    fields['RomSize'] = len(rom)
    lowlevel.SuperMarioHeader.pack_into(rom, 0, **fields)

    # now set the checksums!
    checksum(rom)

    return bytes(rom)



from os import path
import re
import zlib

try:
    from .fast_lzss import compress
except ImportError:
    from .slow_lzss import compress

from . import dispatcher


def append_checksum(binary):
    cksum = ('\r\\ h# %08X' % zlib.adler32(binary)).encode('ascii')
    binary.extend(cksum)


def build(src):
    try:
        with open(path.join(src, 'Bootscript'), 'rb') as f:
            booter = bytearray(f.read().replace(b'\n', b'\r'))
    except (NotADirectoryError, FileNotFoundError):
        raise dispatcher.WrongFormat

    has_checksum = (b'adler32' in booter)

    constants = dict()
    constant_spans = dict()
    for m in re.finditer(rb'h#\s+([A-Fa-f0-9]+)\s+constant\s+([-\w]+)', booter):
        key = m.group(2).decode('ascii')
        val = int(m.group(1), 16)
        constants[key] = val
        constant_spans[key] = m.span(1)

    booter.append(4) # EOT

    for f in ['elf-offset', 'elf-size']:
        assert f in constants

    if 'elf-offset' in constants:
        # special case: pad according to residual info in script
        booter.extend(b'\0' * (constants['elf-offset'] - len(booter)))

        constants['elf-offset'] = len(booter)
        booter.extend(dispatcher.build(path.join(src, 'MacOS.elf')))
        constants['elf-size'] = len(booter) - constants['elf-offset']

    if 'lzss-offset' in constants:
        base = 'lzss'
    else:
        base = 'parcels'

    if base + '-offset' in constants:
        constants[base + '-offset'] = len(booter)
        for attempt in ['MacROM', 'Parcels']:
            try:
                data = dispatcher.build(path.join(src, attempt))
            except:
                pass
            else:
                break
        else:
            raise FileNotFoundError

        if not data.startswith(b'prcl'): data = compress(data)
        booter.extend(data)

        constants[base + '-size'] = len(booter) - constants[base + '-offset']

    constants['info-size'] = len(booter)

    for key, (start, stop) in reversed(sorted(constant_spans.items())):
        insert = ('%X' % constants[key]).zfill(stop - start).encode('ascii')
        assert start + len(insert) == stop
        booter[start:stop] = insert

    if has_checksum: append_checksum(booter)

    return bytes(booter)



# This file is adapted from LZSS.C by Haruhiko Okumura 4/6/1989

# Decompression is pretty quick
# Compression is pretty slow:
# about 50s to compress a 4 MB rom on my machine

from warnings import warn
have_warned_about_slowness = False

N = 0x1000
F = 18
THRESHOLD = 2
NIL = N


def memset(buf, start, stop, to):
    for i in range(start, stop):
        buf[i] = to


def decompress(lzss):
    lzss = iter(lzss)
    plain = bytearray()

    lzdict = bytearray(b' ' * N)

    dict_i = N - F
    def push(byte):
        nonlocal dict_i
        lzdict[dict_i % N] = byte
        dict_i += 1

        plain.append(byte)

    # Iterate through byte-headed "runs"
    try:
        for headerbyte in lzss:
            for bitnum in range(8):
                if (headerbyte >> bitnum) & 1:
                    # Copy a single byte verbatim
                    push(next(lzss))
                else:
                    # Copy 3-18 bytes from the dictionary
                    byte1 = next(lzss)
                    byte2 = next(lzss)
                    lookup_i = (byte2 << 4) & 0xf00 | byte1
                    lookup_len = (byte2 & 0x0f) + 3

                    for i in range(lookup_i, lookup_i+lookup_len):
                        push(lzdict[i % N])

    except StopIteration:
        # Means the last header had <8 real bits, no problem
        pass

    return bytes(plain)


class Tree:
    def __init__(self):
        self.lchild = [0] * (N + 1)
        self.rchild = [0] * (N + 257); memset(self.rchild, N + 1, N + 256 + 1, NIL)
        self.parent = [0] * (N + 1); memset(self.parent, 0, N, NIL)

    # Inserts string of length F, text_buf[r..r+F-1], into one of the trees
    # (text_buf[r]'th tree) and returns the longest-match position and length
    # via the global variables match_position and match_length.
    # If match_length = F, then removes the old node in favor of the new one,
    # because the old one will be deleted sooner. Note r plays double role,
    # as tree node and position in buffer.
    def insert_node(self, r, text_buf):
        lchild, rchild, parent = self.lchild, self.rchild, self.parent

        cmp = 1
        key = text_buf[r:]
        p = N + 1 + key[0]
        rchild[r] = lchild[r] = NIL

        match_length = 0
        match_position = 0

        while 1:
            if cmp >= 0:
                if rchild[p] != NIL:
                    p = rchild[p]
                else:
                    rchild[p] = r
                    parent[r] = p
                    return match_position, match_length
            else:
                if lchild[p] != NIL:
                    p = lchild[p]
                else:
                    lchild[p] = r
                    parent[r] = p
                    return match_position, match_length

            i = 1
            while i < F:
                cmp = key[i] - text_buf[p + i]
                if cmp != 0: break
                i += 1

            if i > match_length:
                match_position = p
                match_length = i
                if match_length >= F: break # out of while loop

        parent[r] = parent[p]
        lchild[r] = lchild[p]
        rchild[r] = rchild[p]
        parent[lchild[p]] = r
        parent[rchild[p]] = r

        if rchild[parent[p]] == p:
            rchild[parent[p]] = r
        else:
            lchild[parent[p]] = r

        parent[p] = NIL;

        return match_position, match_length

    # deletes node p from tree
    def delete_node(self, p):
        lchild, rchild, parent = self.lchild, self.rchild, self.parent

        if parent[p] == NIL: return

        if rchild[p] == NIL:
            q = lchild[p]
        elif lchild[p] == NIL:
            q = rchild[p]
        else:
            q = lchild[p]
            if rchild[q] != NIL:
                while 1:
                    q = rchild[q]
                    if rchild[q] == NIL: break

                rchild[parent[q]] = lchild[q]
                parent[lchild[q]] = parent[q]
                lchild[q] = lchild[p]
                parent[lchild[p]] = q

            rchild[q] = rchild[p]
            parent[rchild[p]] = q

        parent[q] = parent[p]

        if rchild[parent[p]] == p:
            rchild[parent[p]] = q
        else:
            lchild[parent[p]] = q

        parent[p] = NIL


def compress(plain):
    global have_warned_about_slowness

    if not have_warned_about_slowness:
        have_warned_about_slowness = True
        warn('Using slow pure-Python LZSS compression')

    if not plain: return b''

    # Init the variables that get shared with the two closures below
    tree = Tree()
    text_buf = bytearray(N + F - 1); memset(text_buf, 0, N - F, ord(' '))
    match_length = match_position = 0

    # End of function defs, now onto the main attraction
    plain_len = len(plain)
    plain_i = 0

    # code_buf[1..16] saves eight units of code, and code_buf[0] works
    # as eight flags, "1" representing that the unit is an unencoded
    # letter (1 byte), "" a position-and-length pair (2 bytes).
    # Thus, eight units require at most 16 bytes of code.
    code_buf = bytearray(1)
    code_buf_list = [code_buf]
    mask = 1

    # Clear the buffer with any character that will appear often.
    s = 0;  r = N - F

    # Read F bytes into the last F bytes of the buffer
    tblen = 0
    while tblen < F and plain_i < plain_len:
        text_buf[r + tblen] = plain[plain_i]
        tblen += 1
        plain_i += 1

    # Insert the F strings, each of which begins with one or more
    # 'space' characters.  Note the order in which these strings are
    # inserted.  This way, degenerate trees will be less likely to occur.
    for i in range(1, F+1):
        tree.insert_node(r - i, text_buf)

    # Finally, insert the whole string just read.
    # The global variables match_length and match_position are set.
    match_position, match_length = tree.insert_node(r, text_buf)
    while 1:
        match_length = min(match_length, tblen)

        if match_length <= THRESHOLD:
            # Not long enough match.  Send one byte.
            match_length = 1
            code_buf[0] |= mask # 'send one byte' flag
            code_buf.append(text_buf[r]) # Send uncoded.
        else:
            # Send position and length pair. Note match_length > THRESHOLD.
            byte1 = match_position & 0xFF
            byte2 = (match_position >> 4 & 0xF0) | (match_length - THRESHOLD - 1)
            code_buf.append(byte1)
            code_buf.append(byte2)

        # Shift mask left one bit.
        mask = (mask << 1) & 0xFF
        # Send at most 8 units of code together
        if mask == 0:
            code_buf = bytearray(1)
            code_buf_list.append(code_buf)
            mask = 1

        last_match_length = match_length
        i = 0
        while i < last_match_length and plain_i < plain_len:
            tree.delete_node(s) # Delete old strings and
            c = plain[plain_i]; plain_i += 1
            text_buf[s] = c # read new bytes

            # If the position is near the end of buffer, extend the buffer
            # to make string comparison easier.
            if s < F - 1:
                text_buf[s + N] = c

            # Since this is a ring buffer, increment the position modulo N.
            s = (s + 1) % N
            r = (r + 1) % N

            # Register the string in text_buf[r..r+F-1]
            match_position, match_length = tree.insert_node(r, text_buf)

            i += 1

        while i < last_match_length:
            tree.delete_node(s)

            # After the end of text, no need to read,
            s = (s + 1) % N
            r = (r + 1) % N

            # but buffer may not be empty.
            tblen -= 1
            if tblen:
                match_position, match_length = tree.insert_node(r, text_buf)

            i += 1

        # until length of string to be processed is zero
        if tblen == 0: break

    if len(code_buf_list[-1]) == 1:
        code_buf_list.pop()

    return b''.join(code_buf_list)



import struct
from collections import namedtuple

class NamedTupleStruct(struct.Struct):
    """A Struct that works with namedtuple instead of tuple"""

    def __init__(self, *args, name=None, fields=None, **kwargs):
        self.__namedtuple = namedtuple(name, fields)
        self._fields = self.__namedtuple._fields
        super().__init__(*args, **kwargs)

    def __tuplify(self, *args, **kwargs):
        kwargs = {k:v for (k,v) in kwargs.items() if k in self.__namedtuple._fields}
        return self.__namedtuple(*args, **kwargs)

    def unpack(self, *args, **kwargs):
        orig = super().unpack(*args, **kwargs)
        return self.__namedtuple(*orig)

    def unpack_from(self, *args, **kwargs):
        orig = super().unpack_from(*args, **kwargs)
        return self.__namedtuple(*orig)

    def pack(self, *args, **kwargs):
        nt = self.__tuplify(*args, **kwargs)
        return super().pack(*nt)

    def pack_into(self, buf, offset, *args, **kwargs):
        nt = self.__tuplify(*args, **kwargs)
        return super().pack_into(buf, offset, *nt)
